local S = technic.getter

local desc = S("User world anchor\n")

local function compute_forceload_positions(pos, meta)
	local radius = meta:get_int("radius")
	local minpos = vector.subtract(pos, vector.new(radius, radius, radius))
	local maxpos = vector.add(pos, vector.new(radius, radius, radius))
	local minbpos = {}
	local maxbpos = {}
	for _, coord in ipairs({"x","y","z"}) do
		minbpos[coord] = math.floor(minpos[coord] / 16) * 16
		maxbpos[coord] = math.floor(maxpos[coord] / 16) * 16
	end
	local flposes = {}
	for x = minbpos.x, maxbpos.x, 16 do
		for y = minbpos.y, maxbpos.y, 16 do
			for z = minbpos.z, maxbpos.z, 16 do
				table.insert(flposes, vector.new(x, y, z))
			end
		end
	end
	return flposes
end

local function currently_forceloaded_positions(meta)
	local ser = meta:get_string("forceloaded")
	return ser == "" and {} or minetest.deserialize(ser)
end

local function forceload_off(meta)
	local flposes = currently_forceloaded_positions(meta)
	meta:set_string("forceloaded", "")
	for _, p in ipairs(flposes) do
		minetest.forceload_free_block(p)
	end
end

local function forceload_on(pos, meta)
	local want_flposes = compute_forceload_positions(pos, meta)
	local have_flposes = {}
	for _, p in ipairs(want_flposes) do
		if minetest.forceload_block(p) then
			table.insert(have_flposes, p)
		end
	end
	meta:set_string("forceloaded", #have_flposes == 0 and "" or minetest.serialize(have_flposes))
end

local function set_display(pos, meta)
	meta:set_string("infotext", S(meta:get_int("enabled") ~= 0 and "%s Enabled" or "%s Disabled"):format(desc) .. "\n Time left: " .. meta:get_int("time"))
	meta:set_string("formspec",
		"size[8,7.5]"..
		"item_image[2,0;1,1;technic_user_anchor:user_anchor]"..
		"label[0,0;"..minetest.formspec_escape(S("Fuel:")).."]"..
		"label[0,3;"..minetest.formspec_escape(S("Time left: "))..meta:get_int("time").."]"..
		"label[3,0;"..minetest.formspec_escape(desc).."]"..
		"list[current_name;fuel;0,0.5;2,2;]"..
		"list[current_player;main;0,3.5;8,4;]"..
		"label[2,1;"..minetest.formspec_escape(S("Owner:").." "..meta:get_string("owner")).."]"..
		(meta:get_int("locked") == 0 and
			"button[5,1;2,1;lock;"..minetest.formspec_escape(S("Unlocked")).."]" or
			"button[5,1;2,1;unlock;"..minetest.formspec_escape(S("Locked")).."]")..
		"field[2.25,2.3;1,1;radius;"..minetest.formspec_escape(S("Radius:"))..";"..meta:get_int("radius").."]"..
		(meta:get_int("enabled") == 0 and
			"button[5,2;2,1;enable;"..minetest.formspec_escape(S("Disabled")).."]" or
			"button[5,2;2,1;disable;"..minetest.formspec_escape(S("Enabled")).."]")..
		"label[2,3;"..minetest.formspec_escape(S("Keeping %d/%d map blocks loaded"):format(#currently_forceloaded_positions(meta), #compute_forceload_positions(pos, meta))).."]")

end

minetest.register_node("technic_user_anchor:user_anchor", {
	description = desc,
	drawtype = "normal",
	tiles = {"user_anchor_top.png",  "user_anchor_top.png",
	         "user_anchor_side.png", "user_anchor_side.png",
	         "user_anchor_side.png", "user_anchor_side.png"},
	is_ground_content = true,
	groups = {cracky=2},
	sounds = default.node_sound_stone_defaults(),
	allow_metadata_inventory_put = technic.machine_inventory_put,
	on_construct = function(pos)
		local meta = minetest.get_meta(pos)
		local inv = meta:get_inventory()
		inv:set_size("fuel", 4)
	end,
	after_place_node = function (pos, placer)
		local meta = minetest.get_meta(pos)
		if placer and placer:is_player() then
			meta:set_string("owner", placer:get_player_name())
		end
		set_display(pos, meta)
	end,
	can_dig = function (pos, player)
		local meta = minetest.get_meta(pos)
		return meta:get_int("locked") == 0 or (player and player:is_player() and player:get_player_name() == meta:get_string("owner"))
	end,
	on_destruct = function (pos)
		local meta = minetest.get_meta(pos)
		forceload_off(meta)
	end,
	on_receive_fields = function (pos, formname, fields, sender)
		local meta = minetest.get_meta(pos)
		if (meta:get_int("locked") ~= 0 or fields.lock) and
				not (sender and sender:is_player() and
					sender:get_player_name() == meta:get_string("owner")) then
			return
		end
		if fields.unlock then meta:set_int("locked", 0) end
		if fields.lock then meta:set_int("locked", 1) end
		if fields.disable or fields.enable or fields.radius then
			forceload_off(meta)
			if fields.disable then meta:set_int("enabled", 0) end
			if fields.enable and meta:get_int("time")>0 then meta:set_int("enabled", 1) end
			if fields.radius and string.find(fields.radius, "^[0-9]+$") and tonumber(fields.radius) < 16 then meta:set_int("radius", fields.radius) end
			if meta:get_int("enabled") ~= 0 then
				forceload_on(pos, meta)
			end
		end
		set_display(pos, meta)
	end,
})

minetest.register_abm({
	nodenames = {"technic_user_anchor:user_anchor"},
	interval = 1,
	chance = 1,
	action = function(pos, node, active_object_count, active_object_count_wider)
		local meta = minetest.get_meta(pos)
		local inv = meta:get_inventory()
		local current_time = meta:get_int("time")
		local add_burning_time = 0
		for listpos,stack in pairs(inv:get_list("fuel")) do
			if stack ~= nil then
				add_burning_time = add_burning_time + 10 * stack:get_count()
				inv:remove_item("fuel",stack)
			end
		end
		current_time = current_time + add_burning_time
		if current_time > 0 then
			if meta:get_int("enabled") == 1 then
				current_time=current_time - 1
			end
			meta:set_int("time", current_time)
			set_display(pos, meta)
			if current_time==0 then
				meta:set_int("enabled", 0)
				forceload_off(meta)
			end
		end
	end,
})
